"""
This file is part of pyspex

https://github.com/rmvanhees/pyspex.git

Python implementation to create the SPEXone CKD product, or update/read CKD
parameters

Copyright (c) 2019-2021 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
from datetime import datetime
from pathlib import Path
# import argparse

import h5py
import numpy as np

from pyspex import version

# - global parameters ------------------------------
VERSION = '000100'


# - local functions --------------------------------
def str_netcdf():
    """
    return string to identify a netCDF dimension
    """
    return 'This is a netCDF dimension but not a netCDF variable.'


# - class CKDio -------------------------
class CKDio:
    """
    Defines class to read SPEXone CKD parameters

    Attributes
    ----------
    filename :  str
    offset : int
    verbose : bool

    Methods
    -------
    add_offset(values, binning=-1)
       Add Offset CKD to CKD product.
    add_darkcur(values, binning=-1)
       Add Dark-current CKD to CKD product.
    add_non_linearity(values, binning=-1)
       Add non-linearity CKD to CKD product.
    add_prnu(values, binning=-1)
       Add PRNU CKD to CKD product.
    get_offset(binning=-1)
       Read Offset CKD.
    get_darkcur(binning=-1)
       Read Dark-current CKD.
    get_non_linearity(values, binning=-1)
       Read non-Linearity CKD.
    get_prnu(binning=-1)
       Read PRNU CKD.

    Notes
    -----

    Examples
    --------
    """
    def __init__(self, ckd_file, verbose=False):
        """
        Initialize class attributes

        Parameters
        ----------
        ckd_file :  str
           Name of CKD file
        verbose :  bool
           Be verbose
        """
        self.filename = ckd_file
        self.offset = 0
        self.verbose = verbose

    def __initialize_product(self, valid_date=None, overwrite=False):
        """
        Initialize SPEXone product with CKD parameters

        Define name of SPEXone CKD product as:
           SPX1_L1_CKD_<date_valid>_<version>_<date_created>.h5
        - date_valid : ISO UTC date-time as yymmddTHHMMSS
        - version : 6 digits as MMmmrr (for Major, minor and release versions)
        - date_created : ISO UTC date-time as yymmddTHHMMSS
        """
        if valid_date is None:
            valid_date = '20191001T000000'

        creation_date = datetime.utcnow().strftime("%Y%m%dT%H%M%S")
        self.filename = \
            f'SPX1_L1_CKD_{valid_date:13s}_{VERSION:6s}_{creation_date:13s}.h5'

        if Path(self.filename).is_file():
            if not overwrite:
                raise FileExistsError
            Path(self.filename).unlink()

        with h5py.File(self.filename, 'w') as fid:
            fid.attrs['title'] = 'SPEXone Level-1 CKD product'
            fid.attrs['comment'] = ''
            fid.attrs['name'] = self.filename
            fid.attrs['type'] = 'CKD product'
            fid.attrs['mission'] = 'PACE'
            fid.attrs['instrument'] = 'SPEXone'
            fid.attrs['creator'] = \
                'SRON Netherlands Institute for Space Research'
            fid.attrs['creator_date'] = creation_date
            fid.attrs['creator_version'] = VERSION
            fid.attrs['sw_version'] = version.get()
            fid.attrs['valitity_start'] = valid_date
            fid.attrs['validity_stop'] = '20500101T000000'

            grp = fid.create_group('/FULL_FRAME')
            # create dimensions
            dset = grp.create_dataset('row', (2048,), dtype='u2')
            dset.make_scale(str_netcdf())
            dset.attrs['long_name'] = 'row'
            dset[:] = np.arange(0, 2048, dtype='u2')

            dset = grp.create_dataset('column', (2048,), dtype='u2')
            dset.make_scale(str_netcdf())
            dset.attrs['long_name'] = 'column'
            dset[:] = np.arange(0, 2048, dtype='u2')

            _ = fid.create_group('/BINNING_00')
            # create dimensions of default binning scheme

    def add_offset(self, values, binning=-1):
        """
        Add Offset CKD to CKD product
        """
        if self.filename is None or not Path(self.filename).is_file():
            self.__initialize_product()

        with h5py.File(self.filename, 'r+') as fid:
            if binning == -1:
                grp = fid['/FULL_FRAME']
            else:
                grp = fid[f'/BINNING_{binning:03d}']
            if 'offset' in fid:
                raise ValueError('dataset offset already exists')

            # pylint: disable=no-member
            frame_shape = (grp['row'].size, grp['column'].size)
            dset = grp.create_dataset('offset', frame_shape,
                                      compression=1, shuffle=True,
                                      fillvalue=np.nan, dtype='f4')
            dset.dims[0].attach_scale(grp['row'])
            dset.dims[1].attach_scale(grp['column'])
            dset.attrs['long_name'] = "Offset"
            dset[:] = values
            dset.attrs['date_created'] = \
                datetime.utcnow().isoformat(timespec='milliseconds')

    def add_darkcur(self, values, binning=-1):
        """
        Add Dark-current CKD to CKD product
        """
        if self.filename is None or not Path(self.filename).is_file():
            self.__initialize_product()

        with h5py.File(self.filename, 'r+') as fid:
            if binning == -1:
                grp = fid['/FULL_FRAME']
            else:
                grp = fid[f'/BINNING_{binning:03d}']
            if 'darkcurrent' in fid:
                raise ValueError('dataset darkcurrent already exists')

            # pylint: disable=no-member
            frame_shape = (grp['row'].size, grp['column'].size)
            dset = grp.create_dataset('darkcurrent', frame_shape,
                                      compression=1, shuffle=True,
                                      fillvalue=np.nan, dtype='f4')
            dset.dims[0].attach_scale(grp['row'])
            dset.dims[1].attach_scale(grp['column'])
            dset.attrs['long_name'] = "Dark-current"
            dset[:] = values
            dset.attrs['date_created'] = \
                datetime.utcnow().isoformat(timespec='milliseconds')

    def add_non_linearity(self, values, binning=-1):
        """
        Add non-linearity CKD to CKD product

        Notes
        -----
        Not implemented
        """
        if self.filename is None or not Path(self.filename).is_file():
            self.__initialize_product()

        with h5py.File(self.filename, 'r+') as fid:
            if binning == -1:
                grp = fid['/FULL_FRAME']
            else:
                grp = fid[f'/BINNING_{binning:03d}']
            if 'nonlinearity' in fid:
                raise ValueError('dataset nonlinearity already exists')
            raise ValueError('not yet implemented')

    def add_prnu(self, values, binning=-1):
        """
        Add PRNU CKD to CKD product
        """
        if self.filename is None or not Path(self.filename).is_file():
            self.__initialize_product()

        with h5py.File(self.filename, 'r+') as fid:
            if binning == -1:
                grp = fid['/FULL_FRAME']
            else:
                grp = fid[f'/BINNING_{binning:03d}']
            if 'PRNU' in fid:
                raise ValueError('dataset PRNU already exists')

            # pylint: disable=no-member
            frame_shape = (grp['row'].size, grp['column'].size)
            dset = grp.create_dataset('PRNU', frame_shape,
                                      compression=1, shuffle=True,
                                      fillvalue=np.nan, dtype='f4')
            dset.dims[0].attach_scale(grp['row'])
            dset.dims[1].attach_scale(grp['column'])
            dset.attrs['long_name'] = "Pixel Response Non-Uniformity"
            dset[:] = values
            dset.attrs['date_created'] = \
                datetime.utcnow().isoformat(timespec='milliseconds')

    def get_offset(self, binning=-1):
        """
        Read Offset CKD
        """
        with h5py.File(self.filename, 'r') as fid:
            if binning == -1:
                grp = fid['/FULL_FRAME']
            else:
                grp = fid[f'/BINNING_{binning:03d}']
            if 'offset' not in grp:
                raise ValueError('CKD offset does not exist')

            values = grp['offset'][:]

        return values

    def get_darkcur(self, binning=-1):
        """
        Read Dark-current CKD
        """
        with h5py.File(self.filename, 'r') as fid:
            if binning == -1:
                grp = fid['/FULL_FRAME']
            else:
                grp = fid[f'/BINNING_{binning:03d}']
            if 'darkcurrent' not in grp:
                raise ValueError('CKD darkcurrent does not exist')

            values = grp['darkcurrent'][:]

        return values

    def get_non_linearity(self, values, binning=-1):
        """
        Read non-Linearity CKD
        """
        with h5py.File(self.filename, 'r') as fid:
            if binning == -1:
                grp = fid['/FULL_FRAME']
            else:
                grp = fid[f'/BINNING_{binning:03d}']
            if 'nonlinearity' not in grp:
                raise ValueError('CKD nonlinearity does not exist')

            values = grp['nonlinearity'][:]

        return values

    def get_prnu(self, binning=-1):
        """
        Read PRNU CKD
        """
        with h5py.File(self.filename, 'r') as fid:
            if binning == -1:
                grp = fid['/FULL_FRAME']
            else:
                grp = fid[f'/BINNING_{binning:03d}']
            if 'PRNU' not in grp:
                raise ValueError('CKD PRNU does not exist')

            values = grp['PRNU'][:]

        return values


# - main function ----------------------------------
def main():
    """
    main function
    """
    ckd = CKDio(None, verbose=True)
    ckd.add_offset(np.zeros((2048, 2048), dtype=float))
    ckd.add_darkcur(np.zeros((2048, 2048), dtype=float))
    ckd.add_prnu(np.ones((2048, 2048), dtype=float))

    print(np.mean(ckd.get_offset()))
    print(np.mean(ckd.get_darkcur()))
    print(np.mean(ckd.get_prnu()))


# --------------------------------------------------
if __name__ == '__main__':
    main()



#!/usr/bin/env python3
"""
This file is part of pyspex

https://github.com/rmvanhees/pyspex.git

Create or update SPEXone binning-table CKD.
Input is binning table information in CSV format.
The binning-table CKD is written in netCDF4 format.

Copyright (c) 2020-2021 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
import argparse
from pathlib import Path

import numpy as np

from pyspex.binning_tables import BinningTables


# - global parameters ------------------------------


# - local functions --------------------------------


# - main function ----------------------------------
def main():
    """
    main program to build a database with SPEXone binning tables
    """
    parser = argparse.ArgumentParser(
        description='build a database with SPEXone binning tables')
    parser.add_argument('--verbose', default=False, action='store_true',
                        help='be verbose, default be silent')
    parser.add_argument('--validity_start', default=None, type=str,
                        help=('validity start of the binning tables'
                              ' (format: yyyymmddTHHMMSS)'))
    parser.add_argument('--table_id', type=str, default='1',
                        help='start value table_id or comma seperated list')
    parser.add_argument('--lineskip', default=None,
                        help='provide lineskip arrays per binning table')
    parser.add_argument('file_list', nargs='+',
                        help='provide binning table(s) in csv-format')
    args = parser.parse_args()
    if args.verbose:
        print(args)

    # define a table_id per binning table
    table_id_list = args.table_id.split(',')
    if len(table_id_list) == 1:
        offs = int(args.table_id)
        table_id_list = [offs + x for x in range(len(args.file_list))]
    else:
        table_id_list = [int(x) for x in table_id_list]
    if len(table_id_list) != len(args.file_list):
        raise KeyError('you should provide one table_id per binning table')

    # read lineskip data
    lineskip_data = np.loadtxt(args.lineskip, delimiter=',', dtype=np.uint8)
    if lineskip_data.ndim == 1:
        lineskip_data = lineskip_data[np.newaxis, :]
    elif lineskip_data.shape[0] > len(args.file_list):
        lineskip_data = lineskip_data[1:, :]

    # create BinningTable object
    bin_ckd = BinningTables(ckd_dir='.')
    bin_ckd.create_if_needed(args.validity_start)

    # add binning tables
    for ii, flname in enumerate(args.file_list):
        flpath = Path(flname)
        if not flpath.is_file():
            raise FileNotFoundError(flname)
        if flpath.suffix != '.csv':
            continue

        # read csv-file
        table = np.loadtxt(flpath, delimiter=',', dtype=np.uint32)

        # write new binning-table
        bin_ckd.add_table(table_id_list[ii], lineskip_data[ii, :], table)


# --------------------------------------------------
if __name__ == '__main__':
    main()



#!/usr/bin/env python3
"""
This file is part of pyspex

https://github.com/rmvanhees/pyspex.git

Python implementation SPEXone instrument simulator output to L1A

Environment
-----------
CKD_DIR :  directory with SPEXone CKD, default is CWD

Copyright (c) 2019-2021 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
import argparse
from datetime import datetime, timezone
from os import environ
from pathlib import Path

import h5py
import numpy as np
import xarray as xr

from pyspex import spx_product
from pyspex.lib.tmtc_def import tmtc_def
from pyspex.tif_io import TIFio
from pyspex.lv1_io import L1Aio
from pyspex.lv1_gse import LV1gse

# - global parameters ------------------------------
EPOCH = datetime(1970, 1, 1, tzinfo=timezone.utc)


# - local functions --------------------------------
def get_table_id(ckd_dir, bin_table_name):
    """
    Quick en dirty implementation to obtain table_id from binning-table CKD
    """
    if not Path(ckd_dir).is_dir():
        ckd_dir = environ.get('CKD_DIR', '.')

    # only read the latest version of the binning-table CKD
    bin_tables = sorted(list(Path(ckd_dir).glob('SPX1_OCAL_L1A_TBL_*.nc')))
    if not bin_tables:
        raise FileNotFoundError('No binning table found')

    table_id = None
    with h5py.File(bin_tables[-1], 'r') as fid:
        for grp in [x for x in fid.keys() if x.startswith('Table')]:
            gid = fid[grp]
            if 'origin' not in gid.attrs:
                continue

            if gid.attrs['origin'].decode('ascii') == bin_table_name:
                table_id = int(grp.split('_')[1])
                break

    return table_id


def get_stimulus(hdr):
    """
    Return stimulus data as a xarray::Dataset
    """
    if 'Spectral data stimulus' not in hdr:
        return None

    ds_dict = hdr['Spectral data stimulus']
    xr_wv = xr.DataArray(ds_dict['Wavelength (nm)'],
                         coords=[ds_dict['Wavelength (nm)']],
                         dims=['wavelength'],
                         attrs={'long_name': 'wavelength of stimulus',
                                'units': 'nm'})

    data = ds_dict['Radiance (photons/(s.nm.m^2.sr)']
    data *= 2.99792458 * 6.62607015e-14 / ds_dict['Wavelength (nm)']
    xr_sign = xr.DataArray(data, dims=['wavelength'],
                           coords=[ds_dict['Wavelength (nm)']],
                           attrs={'long_name': 'signal of stimulus',
                                  'units': 'W.m-2.sr-1.um-1'})

    return xr.Dataset({'wavelength': xr_wv, 'signal': xr_sign})


# - main function ----------------------------------
def main():
    """
    main program to illustate the creation of a L1A calibration product
    """
    parser = argparse.ArgumentParser(
        description=('create SPEXone L1A product from instrument simulations'))
    parser.add_argument('--lineskip', default=False, action='store_true',
                        help='read line-skip data, when available')
    parser.add_argument('--output', default=None,
                        help='define output directory, default=CWD')
    inp_grp = parser.add_argument_group('inp_grp', 'process inp_tif files')
    inp_grp.add_argument('--inp_tif', default=False, action='store_true',
                         help='use TIFF file with 32-bit image')
    inp_grp.add_argument('--nframe', default=None, type=int,
                         help='specify the number of images to be generated')
    parser.add_argument('--verbose', default=False, action='store_true',
                        help='be verbose')
    parser.add_argument('ascii_file', default=None,
                        help=('provide path to SPEXone instrument simulator'
                              ' product with extension ".dat"'))
    args = parser.parse_args()
    if args.verbose:
        print(args)

    if not Path(args.ascii_file).is_file():
        raise FileNotFoundError(f'File {args.ascii_file} does not exist')

    if not args.inp_tif and args.nframe is not None:
        raise RuntimeError('option --nframe works only with "--inp_tif"')

    tif = TIFio(args.ascii_file, inp_tif=args.inp_tif, lineskip=args.lineskip)
    hdr = tif.header()
    tags = tif.tags()[0]
    images = tif.images(n_frame=args.nframe)
    if args.verbose:
        for key, value in hdr.items():
            if key == 'Spectral data stimulus':
                continue
            print(f"{key:s}: '{value}'")

    # For now, we obtain the reference time from the TIFF.tags
    msm_id = hdr['OCAL measurement']
    for key in tags:
        if key.name == 'datetime':
            str_date, str_time = tags[key.value].split(' ')
            utc_start = datetime.fromisoformat(
                str_date.replace(':', '-') + 'T' + str_time + '+00:00')
            break

    prod_name = spx_product.prod_name(utc_start, msm_id=msm_id)
    if args.inp_tif:
        prod_name = prod_name.replace('_L1A_', '_inp_L1A_')
    if args.output is not None:
        dest_dir = Path(args.output)
        if not dest_dir.is_dir():
            dest_dir.mkdir()
        prod_name = str(dest_dir / prod_name)
    if args.verbose:
        print(prod_name)

    # define dimensions for L1A product
    if args.inp_tif:
        if images.ndim == 2:
            n_images = 1
            n_samples = images.size
        else:
            n_images = images.shape[0]
            n_samples = images.shape[1] * images.shape[2]
    else:
        n_images = int(hdr['Number of measurements'])
        if args.lineskip:
            n_samples = (int(hdr['Enabled lines'])
                         * int(hdr['Column dimension']))
        else:
            n_samples = (int(hdr['Row dimension'])
                         * int(hdr['Column dimension']))

    dims = {'number_of_images': n_images,
            'samples_per_image': n_samples,
            'hk_packets': n_images,
            'viewing_angles': int(hdr['Number of viewing angles'])}

    # define binning table ID
    if n_samples == 0x400000:
        table_id = 0
    elif n_samples == 0x100000:
        table_id = 1
    else:
        if args.lineskip:
            table_id = get_table_id('/nfs/SPEXone/share/ckd',
                                    hdr['Line-enable-array'])
        else:
            table_id = get_table_id('/nfs/SPEXone/share/ckd',
                                    hdr['Flexible binning table'])

    # Compute delta_time for each frame (seconds)
    # convert timestamps to seconds per day
    offs = (utc_start - EPOCH).total_seconds()
    intg_time = int(hdr['Co-additions']) * float(hdr['Exposure time (s)'])
    secnds = (1 + np.arange(n_images, dtype=float)) * intg_time
    img_sec = (offs + secnds // 1).astype('u4')
    img_subsec = np.round((secnds % 1) * 2**16).astype('u2')

    nom_hk = np.zeros(n_images, dtype=np.dtype(tmtc_def(0x320)))
    sci_hk = np.zeros(n_images, dtype=np.dtype(tmtc_def(0x350)))
    if table_id == 0:
        sci_hk['REG_FULL_FRAME'] = 1
        sci_hk['REG_CMV_OUTPUTMODE'] = 3
    else:
        sci_hk['REG_FULL_FRAME'] = 2
        sci_hk['REG_CMV_OUTPUTMODE'] = 1
    sci_hk['REG_BINNING_TABLE_START'] = 0x400000 * (table_id + 1) + 0x80000000
    sci_hk['DET_OFFSET'] = 16384 - 70
    sci_hk['DET_FOTLEN'] = 20
    sci_hk['DET_EXPTIME'] = \
        int(1e7 * float(hdr['Exposure time (s)']) / 129 - 0.43 * 20)
    sci_hk['REG_NCOADDFRAMES'] = int(hdr['Co-additions'])

    # Generate L1A product
    with L1Aio(prod_name, dims=dims) as l1a:
        # write image data, detector telemetry and image attributes
        l1a.fill_science(images.reshape(n_images, -1), sci_hk,
                         np.arange(n_images))
        l1a.fill_time(img_sec, img_subsec, group='image_attributes')

        # Engineering data
        l1a.fill_nomhk(nom_hk)
        l1a.fill_time(img_sec, img_subsec, group='engineering_data')

        # Global attributes
        l1a.fill_global_attrs(inflight=False)
        l1a.set_attr('input_files', [Path(args.ascii_file).name])

    with LV1gse(prod_name) as gse:
        gse.set_attr('origin', hdr['history'])
        gse.set_attr('comment', hdr['OCAL measurement'])
        if 'Light source' in hdr:
            gse.set_attr('light_source', hdr['Light source'])
        if 'Detector illumination' in hdr:
            gse.set_attr('Light_source', hdr['Detector illumination'])
            gse.set_attr('Illumination_level',
                         float(hdr['Detect. illumination e/ms']))
        if 'Detector temperature (K)' in hdr:
            gse.set_attr('Detector_temperature',
                         float(hdr['Detector temperature (K)']))
        if 'Detector temperature (K)' in hdr:
            gse.set_attr('Optics_temperature',
                         float(hdr['Optics temperature (K)']))
        if 'Illuminated viewing angle' in hdr:
            vp_dict = {'0': 1, '1': 2, '2': 16, '3': 8, '4': 4}
            gse.write_viewport(vp_dict.get(hdr['Illuminated viewing angle']))
        else:
            gse.write_viewport(0)
        if 'Illuminated field ACT' in hdr:
            gse.write_attr_act(float(hdr['Rotation stage ACT angle']),
                               float(hdr['Illuminated field ACT']))
        else:
            gse.write_attr_act(float(hdr['Rotation stage ACT angle']))
        if 'Illuminated field ALT' in hdr:
            gse.write_attr_alt(float(hdr['Rotation stage ALT angle']),
                               float(hdr['Illuminated field ALT']))
        else:
            gse.write_attr_alt(float(hdr['Rotation stage ALT angle']))
        if 'DoLP input' in hdr:
            gse.write_attr_polarization(float(hdr['AoLP input (deg)']),
                                        float(hdr['DoLP input']))
        if 'Spectral data stimulus' in hdr \
           and 'Illuminated viewing angle' in hdr:
            gse.write_data_stimulus(get_stimulus(hdr))


# --------------------------------------------------
if __name__ == '__main__':
    main()



"""
This file is part of pyspex

https://github.com/rmvanhees/pyspex.git

Class to convert TAI to UTC and visa versa

Copyright (c) 2022 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
from datetime import datetime

# - global parameters ------------------------------
EPOCH_1900 = int(datetime.fromisoformat("1900-01-01T00+00:00").timestamp())
EPOCH_1958 = int(datetime.fromisoformat("1958-01-01T00+00:00").timestamp())


# - class Clocks -------------------------
class Clocks:
    """
    Convert TAI to UTC or visa versa
    """
    def __init__(self):
        """
        Initializes class Clocks with table to convert TAI to UTC

        Notes
        -----
        Information obtained from IERS Bulletin C62

        origin: https://www.ietf.org/timezones/data/leap-seconds.list

        Valid until:  28 June 2022
        """
        # This table contains 3 columns:
        #  1) epoch as a number of seconds since 1 January 1900, 00:00:00
        #  2) number of seconds that must be added to UTC to compute TAI
        #  3) UTC time corresponding to the epoch in the first column
        self.table = [
            (2272060800, 10, "1972-01-01T00+00:00"),    # 1 Jan 1972
            (2287785600, 11, "1972-07-01T00+00:00"),    # 1 Jul 1972
            (2303683200, 12, "1973-01-01T00+00:00"),    # 1 Jan 1973
            (2335219200, 13, "1974-01-01T00+00:00"),    # 1 Jan 1974
            (2366755200, 14, "1975-01-01T00+00:00"),    # 1 Jan 1975
            (2398291200, 15, "1976-01-01T00+00:00"),    # 1 Jan 1976
            (2429913600, 16, "1977-01-01T00+00:00"),    # 1 Jan 1977
            (2461449600, 17, "1978-01-01T00+00:00"),    # 1 Jan 1978
            (2492985600, 18, "1979-01-01T00+00:00"),    # 1 Jan 1979
            (2524521600, 19, "1980-01-01T00+00:00"),    # 1 Jan 1980
            (2571782400, 20, "1981-07-01T00+00:00"),    # 1 Jul 1981
            (2603318400, 21, "1982-07-01T00+00:00"),    # 1 Jul 1982
            (2634854400, 22, "1983-07-01T00+00:00"),    # 1 Jul 1983
            (2698012800, 23, "1985-07-01T00+00:00"),    # 1 Jul 1985
            (2776982400, 24, "1988-01-01T00+00:00"),    # 1 Jan 1988
            (2840140800, 25, "1990-01-01T00+00:00"),    # 1 Jan 1990
            (2871676800, 26, "1991-01-01T00+00:00"),    # 1 Jan 1991
            (2918937600, 27, "1992-07-01T00+00:00"),    # 1 Jul 1992
            (2950473600, 28, "1993-07-01T00+00:00"),    # 1 Jul 1993
            (2982009600, 29, "1994-07-01T00+00:00"),    # 1 Jul 1994
            (3029443200, 30, "1996-01-01T00+00:00"),    # 1 Jan 1996
            (3076704000, 31, "1997-07-01T00+00:00"),    # 1 Jul 1997
            (3124137600, 32, "1999-01-01T00+00:00"),    # 1 Jan 1999
            (3345062400, 33, "2006-01-01T00+00:00"),    # 1 Jan 2006
            (3439756800, 34, "2009-01-01T00+00:00"),    # 1 Jan 2009
            (3550089600, 35, "2012-07-01T00+00:00"),    # 1 Jul 2012
            (3644697600, 36, "2015-07-01T00+00:00"),    # 1 Jul 2015
            (3692217600, 37, "2017-01-01T00+00:00"),    # 1 Jan 2017
            (3865363200, None, "2022-06-28T00+00:00")]  # expiring date

        # use epoch since 1970.0 instead 1900.0
        self.table = [(x + EPOCH_1900, y, z) for x, y, z in self.table]

    def to_tai(self, timestamp: int) -> int:
        """
        Return TAI timestamp for given UTC timestamp

        Parameters
        ----------
        timestamp :  int
           timestamp in UTC (epoch 1970.0)

        Return
        ------
        int :  TAI timestamp (epoch 1958.0)
        """
        for iers_epoch, iers_offs, _ in self.table[::-1]:
            if timestamp >= iers_epoch:
                break

        if iers_offs is None:
            raise ValueError('update your leap second list through'
                             ' IERS Bulletin C62')

        return timestamp + iers_offs - EPOCH_1958

    def to_utc(self, timestamp: int) -> int:
        """
        Return UTC timestamp for given TAI timestamp

        Parameters
        ----------
        timestamp :  int
           timestamp in UTC (epoch 1958.0)

        Return
        ------
        int :  UTC timestamp (epoch 1970.0)
        """
        timestamp += EPOCH_1958
        for iers_epoch, iers_offs, _ in self.table[::-1]:
            if timestamp >= iers_epoch:
                break

        if iers_offs is None:
            raise ValueError('update your leap second list through'
                             ' IERS Bulletin C62')

        return timestamp - iers_offs

    def test(self) -> None:
        """
        Perform unit test on class Clocks
        """
        for res in self.table:
            dt_str = datetime.fromisoformat(res[2])
            if res[0] != int(dt_str.timestamp()):
                raise ValueError(f'Detected error in table at epoch={res[0]}')

        print('INFO: current information on leap seconds expires'
              f' at {self.table[-1][2]}')


def main():
    """
    main function
    """
    clocks = Clocks()
    clocks.test()

    for xx in [489023999, 489024000, 489024001]:
        print(f'{xx}'
              f' -> {datetime.utcfromtimestamp(xx).isoformat()}'
              f' -> {clocks.to_tai(xx)}')

    for xx in [1120176021, 1120176022, 1120176023, 1120176024]:
        print(f'{xx} -> {clocks.to_utc(xx)}')

    # check timestamp after the leap seconds file has expires
    timestamp = datetime.fromisoformat('2022-06-28T00+00:00').timestamp()
    clocks.to_tai(timestamp)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
"""
This file is part of pyspex

https://github.com/rmvanhees/pyspex.git

Create new SPEXone Level-1A product with selected data from original

Copyright (c) 2021 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
import argparse
from datetime import datetime, timezone
from pathlib import Path

import h5py
import numpy as np

from pyspex.lv1_io import L1Aio


# --------------------------------------------------
def inv_sec_of_day(reference_day, sec_of_day,
                   epoch=datetime(1970, 1, 1, tzinfo=timezone.utc)):
    """
    Convert seconds after midnight to CCSDS timestamps

    Parameters
    ----------
    reference_day :  datetime object
    sec_of_day :  ndarray
    epoch :  datetime object, optional

    Returns
    -------
    tuple holding CCSDS timestamps
    """
    offs = np.uint32((reference_day - epoch).total_seconds())
    ccsds_sec = sec_of_day.astype('u4')
    ccsds_subsec = (65536 * (sec_of_day - ccsds_sec)).astype('u2')

    return offs + ccsds_sec, ccsds_subsec


# --------------------------------------------------
def main():
    """
    Main function of this module
    """
    parser = argparse.ArgumentParser(
        description=('Copy selected data from one SPEXone L1A product'
                     ' into a new SPEXone L1A product'))
    parser.add_argument('--verbose', '-v', action='store_true',
                        help='be verbose, default be silent')
    parser.add_argument('--mps_id', nargs='*', type=int, default=None,
                        help='select on MPS-ID [comma separated?]')
    parser.add_argument('--mon_type', default=None,
                        help=('Specify monitoring type identifier: '
                              'MON-DARK, MON-NOISE, MON-NLIN, ...')
    # parser.add_argument('--time', nargs=2, default=None,
    #                    help='select on image time [start, end]')
    # parser.add_argument('--', default=None, help='')
    parser.add_argument('--out', default='.',
                        help=('name of directory to store the new Level-1A'
                              ' product, default: current working directory'))
    parser.add_argument('l1a_product', default=None,
                        help='name of SPEXone Level-1A product')
    args = parser.parse_args()
    if args.verbose:
        print(args)

    l1a_product = Path(args.l1a_product)
    if not l1a_product.is_file():
        raise FileNotFoundError(f'File {args.l1a_product} does not exist')
    # ToDo: check if SPEXone Level-1a product
    # ToDo: implement check on data product

    out_dir = Path(args.out)
    if not out_dir.is_dir():
        out_dir.mkdir(mode=0o755, parents=True)

    # ----- read data from orignal product -----
    # pylint: disable=no-member, unsubscriptable-object
    with h5py.File(l1a_product) as fid:
        # read image data, detector telemetry and image attributes
        # - datasets: img_data, img_hk, img_id, img_sec, img_subsec
        img_sec = fid['/image_attributes/image_CCSDS_sec'][:]
        img_subsec = fid['/image_attributes/image_CCSDS_subsec'][:]
        # obtain reference date
        units = fid['/image_attributes/image_time'].attrs['units']
        reference_day = datetime.fromisoformat(units[14:].decode('ascii'))
        sec_of_day = fid['/image_attributes/image_time'][:]
        img_sec, img_subsec = inv_sec_of_day(reference_day, sec_of_day)
        img_id = fid['/image_attributes/image_ID'][:]
        img_data = fid['/science_data/detector_images'][:]
        img_hk = fid['/science_data/detector_telemetry'][:]

        # read engineering data
        # - datasets: nomhk_data, nomhk_sec, nomhk_subsec
        # - datasets: demhk_data
        nomhk_data = fid['/engineering_data/NomHK_telemetry'][:]
        if nomhk_data.size > 0:
            # obtain reference date
            units = fid['/engineering_data/HK_tlm_time'].attrs['units']
            print(units)
            reference_day = datetime.fromisoformat(units[14:].decode('ascii'))
            print(reference_day)
            sec_of_day = fid['/engineering_data/HK_tlm_time'][:]
            nomhk_sec, nomhk_subsec = inv_sec_of_day(reference_day, sec_of_day)

        demhk_data = fid['/engineering_data/DemHK_telemetry'][:]

        # read additional attributes:
        # - inflight
        # - EGSE/OGSE, ...
        inflight = not fid.attrs['institution'].startswith(b'SRON')
        # gse_data = 'gse_data' in fid

    # define dimensions
    dims = {'number_of_images': img_data.shape[0],
            'samples_per_image': img_data.shape[1],
            'hk_packets': nomhk_data.size}

    # ----- perform data selection -----
    # ToDo: implement data selection

    # ----- now we can update the name of the output product -----
    # - because the production time has changed
    # - and when coverage time is changed
    if (out_dir / l1a_product.name).is_file() \
       and l1a_product.samefile(out_dir / l1a_product.name):
        raise OSError('Output will overwrite original product')

    # ----- write new output product with selected data -----
    with L1Aio(out_dir / l1a_product.name, dims=dims) as l1a:
        # write image data, detector telemetry and image attributes
        l1a.fill_science(img_data, img_hk, img_id)
        l1a.fill_time(img_sec, img_subsec, group='image_attributes')

        # write engineering data
        if nomhk_data.size > 0:
            l1a.fill_nomhk(nomhk_data)
            l1a.fill_time(nomhk_sec, nomhk_subsec, group='engineering_data')

        if demhk_data.size > 0:
            l1a.fill_demhk(demhk_data)

        # write global attributes
        l1a.fill_global_attrs(inflight=inflight)
        # l1a.set_attr('input_files', [Path(x).name for x in args.file_list])

    # copy group with EGSE/OGSE data
    # *** DO NOT USE: BREAKS NETCDF4 FORMAT ***
    # if gse_data:
    #     print('copy EGSE/OGSE data')
    #     with h5py.File(l1a_product) as fid_in:
    #         with h5py.File(out_dir / l1a_product.name, 'r+') as fid_out:
    #             fid_out.copy(fid_in['gse_data'], 'gse_data')


# --------------------------------------------------
if __name__ == '__main__':
    main()



#!/usr/bin/env python3
"""
This file is part of pyspex

https://github.com/rmvanhees/pyspex.git

Create new SPEXone Level-1C product with selected data from original

Copyright (c) 2021 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
import argparse
from datetime import datetime
from pathlib import Path

import h5py

from pyspex.lv1_io import L1Cio


# --------------------------------------------------
def main():
    """
    Main function of this module
    """
    parser = argparse.ArgumentParser(
        description=('Copy selected data from one SPEXone L1C product'
                     ' into a new SPEXone L1C product'))
    parser.add_argument('--verbose', '-v', action='store_true',
                        help='be verbose, default be silent')
    parser.add_argument('--mps_id', nargs='*', type=int, default=None,
                        help='select on MPS-ID')
    # parser.add_argument('--time', nargs=2, default=None,
    #                    help='select on image time [start, end]')
    # parser.add_argument('--', default=None, help='')
    parser.add_argument('--out', default='.',
                        help=('name of directory to store the new Level-1C'
                              ' product, default: current working directory'))
    parser.add_argument('l1c_products', default=None, nargs='+',
                        help=('names of SPEXone/OCI Level-1C products'
                              'expect one SPEXone L1C product'
                              ' and one or more OCI L1C granuals'))
    args = parser.parse_args()
    if args.verbose:
        print(args)

    for name in args.l1c_products:
        l1c_product = Path(name)
        if not l1c_product.is_file():
            raise FileNotFoundError(f'File {name} does not exist')
        # ToDo: check SPEXone/OCI products
        # ToDo: store in variables spx1_product:str, oci_products:list
        # ToDo: implement check on data product

    out_dir = Path(args.out_dir)
    if not out_dir.is_dir():
        out_dir.mkdir(mode=0o755, parents=True)

    # ----- read data from orignal product -----
    # ToDo: read data into dict or xarray.Dataset per product
    # ToDo: combine OCI granuals into one dict or xarray.Dataset
    # pylint: disable=no-member, unsubscriptable-object
    with h5py.File(l1c_product[0]) as fid:
        # group BIN_ATTRIBUTES
        ref_date = None
        nadir_view_time = fid['/bin_attributes/nadir_view_time'][:]
        units = fid['/bin_attributes/nadir_view_time'].attrs['units']
        if len(units) > 14:
            ref_date = datetime.fromisoformat(units[14:])
        elif fid.attrs['time_coverage_start'] != b'Unknown':
            ref_date = datetime.fromisoformat(
                fid.attrs['time_coverage_start'].decode('ascii')[:-1])
            ref_date = ref_date.replace(hour=0, minute=0, second=0)
        view_time_offs = fid['/bin_attributes/view_time_offsets'][:]

        # group GEOLOCATION_DATA
        altitude = fid['/geolocation_data/altitude'][:]
        altitude_var = fid['/geolocation_data/altitude_variability'][:]
        latitude = fid['/geolocation_data/latitude'][:]
        longitude = fid['/geolocation_data/longitude'][:]
        sensor_azi = fid['/geolocation_data/sensor_azimuth'][:]
        sensor_zen = fid['/geolocation_data/sensor_zenith'][:]
        solar_azi = fid['/geolocation_data/solar_azimuth'][:]
        solar_zen = fid['/geolocation_data/solar_zenith'][:]

        # group OBSERVATION_DATA
        obs_per_view = fid['/observation_data/obs_per_view'][:]
        qc_bits = fid['/observation_data/QC_bitwise'][:]
        qc_val = fid['/observation_data/QC'][:]
        qc_pol_bits = fid['/observation_data/QC_polsample_bitwise'][:]
        qc_pol = fid['/observation_data/QC_polsample'][:]
        intens_val = fid['/observation_data/I'][:]
        intens_noise = fid['/observation_data/I_noise'][:]
        intens_pol = fid['/observation_data/I_polsample'][:]
        intens_pol_noise = fid['/observation_data/I_polsample_noise'][:]
        aolp_val = fid['/observation_data/AOLP'][:]
        aolp_noise = fid['/observation_data/AOLP_noise'][:]
        dolp_val = fid['/observation_data/DOLP'][:]
        dolp_noise = fid['/observation_data/DOLP_noise'][:]
        q_val = fid['/observation_data/Q_over_I'][:]
        q_noise = fid['/observation_data/Q_over_I_noise'][:]
        u_val = fid['/observation_data/U_over_I'][:]
        u_noise = fid['/observation_data/U_over_I_noise'][:]

        # group SENSOR_VIEWS_BANDS
        intens_bands = fid['/sensor_views_bands/intensity_bandpasses'][:]
        intens_wav = fid['/sensor_views_bands/intensity_wavelengths'][:]
        intens_f0 = fid['/sensor_views_bands/intensity_F0'][:]
        polar_bands = fid['/sensor_views_bands/polarization_bandpasses'][:]
        polar_wav = fid['/sensor_views_bands/polarization_wavelengths'][:]
        polar_f0 = fid['/sensor_views_bands/polarization_F0'][:]
        view_angles = fid['/sensor_views_bands/view_angles'][:]

        # global attributes
        global_attrs = {}
        check_attrs = ("history", "sun_earth_distance",
                       "terrain_data_source",
                       "spectral_response_function",
                       "systematic_uncertainty_model",
                       "bin_size_at_nadir")  # "nadir_bin"
        for key in sorted(check_attrs):
            if key in fid.attrs:
                if isinstance(fid.attrs[key], h5py.Empty):
                    global_attrs[key] = ""
                else:
                    global_attrs[key] = fid.attrs[key].decode('ascii')

    dims = {'bins_along_track':  latitude.shape[0],
            'bins_across_track':  latitude.shape[1],
            'intensity_bands_per_view':  intens_bands.shape[1],
            'polarization_bands_per_view':  polar_bands.shape[1]}

    # ----- perform data selection -----
    # ToDo: implement data selection

    # ----- now we can update the name of the output product -----
    # - because the production time has changed
    # - and when coverage time is changed
    if (out_dir / l1c_product.name).is_file() \
       and l1c_product.samefile(out_dir / l1c_product.name):
        raise OSError('Output will overwrite original product')

    # ----- write new output product with selected data -----
    print('dims: ', dims)
    with L1Cio(out_dir / l1c_product.name, ref_date=ref_date, dims=dims) as l1c:
        # group BIN_ATTRIBUTES
        l1c.set_dset('/BIN_ATTRIBUTES/nadir_view_time', nadir_view_time)
        l1c.set_dset('/BIN_ATTRIBUTES/view_time_offsets', view_time_offs)

        # group GEOLOCATION_DATA
        l1c.set_dset('/GEOLOCATION_DATA/altitude', altitude)
        l1c.set_dset('/GEOLOCATION_DATA/altitude_variability', altitude_var)
        l1c.set_dset('/GEOLOCATION_DATA/latitude', latitude)
        l1c.set_dset('/GEOLOCATION_DATA/longitude', longitude)
        l1c.set_dset('/GEOLOCATION_DATA/sensor_azimuth', sensor_azi)
        l1c.set_dset('/GEOLOCATION_DATA/sensor_zenith', sensor_zen)
        l1c.set_dset('/GEOLOCATION_DATA/solar_azimuth', solar_azi)
        l1c.set_dset('/GEOLOCATION_DATA/solar_zenith', solar_zen)

        # group OBSERVATION_DATA
        l1c.set_dset('/OBSERVATION_DATA/obs_per_view', obs_per_view)
        l1c.set_dset('/OBSERVATION_DATA/QC_bitwise', qc_bits.astype('u4'))
        l1c.set_dset('/OBSERVATION_DATA/QC', qc_val)
        l1c.set_dset('/OBSERVATION_DATA/QC_polsample_bitwise',
                     qc_pol_bits.astype('u4'))
        l1c.set_dset('/OBSERVATION_DATA/QC_polsample', qc_pol)
        l1c.set_dset('/OBSERVATION_DATA/I', intens_val)
        l1c.set_dset('/OBSERVATION_DATA/I_noise', intens_noise)
        l1c.set_dset('/OBSERVATION_DATA/I_polsample', intens_pol)
        l1c.set_dset('/OBSERVATION_DATA/I_polsample_noise', intens_pol_noise)
        l1c.set_dset('/OBSERVATION_DATA/AoLP', aolp_val)
        l1c.set_dset('/OBSERVATION_DATA/AoLP_noise', aolp_noise)
        l1c.set_dset('/OBSERVATION_DATA/DoLP', dolp_val)
        l1c.set_dset('/OBSERVATION_DATA/DoLP_noise', dolp_noise)
        l1c.set_dset('/OBSERVATION_DATA/Q_over_I', q_val)
        l1c.set_dset('/OBSERVATION_DATA/Q_over_I_noise', q_noise)
        l1c.set_dset('/OBSERVATION_DATA/U_over_I', u_val)
        l1c.set_dset('/OBSERVATION_DATA/U_over_I_noise', u_noise)

        # group SENSOR_VIEWS_BANDS
        l1c.set_dset('/SENSOR_VIEWS_BANDS/intensity_bandpasses', intens_bands)
        l1c.set_dset('/SENSOR_VIEWS_BANDS/intensity_wavelengths', intens_wav)
        l1c.set_dset('/SENSOR_VIEWS_BANDS/intensity_F0', intens_f0)
        l1c.set_dset('/SENSOR_VIEWS_BANDS/polarization_bandpasses', polar_bands)
        l1c.set_dset('/SENSOR_VIEWS_BANDS/polarization_wavelengths', polar_wav)
        l1c.set_dset('/SENSOR_VIEWS_BANDS/polarization_F0', polar_f0)
        l1c.set_dset('/SENSOR_VIEWS_BANDS/view_angles', view_angles)

        # write global attributes
        l1c.fill_global_attrs(inflight=True)
        for key, value in global_attrs.items():
            l1c.set_attr(key, value)


# --------------------------------------------------
if __name__ == '__main__':
    main()



"""
This file is part of pyspex

https://github.com/rmvanhees/pyspex.git

Python class to handle SPEXone binning tables

Environment
-----------
CKD_DIR :  directory with SPEXone CKD, default is CWD

Copyright (c) 2021 SRON - Netherlands Institute for Space Research
   All Rights Reserved

License:  BSD-3-Clause
"""
from datetime import datetime, timezone
from os import environ
from pathlib import Path

import numpy as np
from netCDF4 import Dataset

from pyspex import version


# - global parameters ------------------------------
FILL_VALUE = 0xFFFFFFFF  # 0X7FFFFFFF


# - local functions --------------------------------
class BinningTables:
    """
    Defines class to store and obtain SPEXone binning CKD

    Attributes
    ----------
    ckd_dir : pathlib.Path
       Directory with binning table CKD
    ckd_file : str
       Filename of the binning table CKD

    Methods
    -------
    create_if_needed(validity_start, release=1)
       Initialize CKD file with binning tables (netCDF4 format).
    add_table(table_id, lineskip_arr, binning_table)
       Add a binning table definition to existing CKD file.
    search(coverage_start)
       Search current CKD file with binning tables.
    unbin(table_id: int, img_binned)
       Return unbinned detector data.

    Notes
    -----
    The name of the binning-table CKD files is defined as follows:

          SPX1_CKD_BIN_TBL_<yyyymmddTHHMMSS>_<NNN>.nc

    where yyyymmddTHHMMSS defines the validity start (UTC) and NNN the
    release number of the file format.

    The binning tables as defined on-ground are supposed to be available
    during the whole mission at the same memory location. Because these
    original binning tables are necessary for re-processing and may
    facilitate instrument performance monitoring. Therefore, it is prefered
    that new binning tables are added to the current CKD product, without
    changing the validity start string. In case any of the binning tables
    are overwritten or moved in memory, a new CKD product should be released.

    If the format of the CKD changes then new CKD files should be created with
    the release number should be increased by one.

    Examples
    --------
    # create new binning_table CKD
    >>> bin_tbl = BinningTables()
    >>> bin_tbl.create_if_needed(validity_start)
    >>> bin_tbl.add_table(0, lineskip_arr, binning_table)
    >>> bin_tbl.add_table(1, lineskip_arr, binning_table)

    # add new binning-table to existing CKD
    >>> bin_tbl BinningTables()
    >>> bin_tbl.create_if_needed(validity_start)
    >>> bin_tbl.add_table(2, lineskip_arr, binning_table)

    # use binning-table to unbin SPEXone detector data
    >>> bin_tbl BinningTables()
    >>> bin_tbl.search(coverage_start)
    >>> img = bin_tbl.unbin(2, img_binned)
    """
    def __init__(self, ckd_dir=None) -> None:
        """
        Initialize class attributes

        Parameters
        ----------
        ckd_dir : str
           Name of directory for the binning table CKD

        Raises
        ------
        FileNotFoundError
           directory with SPEXone CKD does not exist
        """
        if ckd_dir is None:
            self.ckd_dir = Path('/nfs/SPEXone/share/ckd')
            if not self.ckd_dir.is_dir():
                self.ckd_dir = Path(environ.get('CKD_DIR', '.'))
        else:
            self.ckd_dir = Path(ckd_dir)
        if not self.ckd_dir.is_dir():
            raise FileNotFoundError('directory with SPEXone CKD does not exist')

        self.ckd_file = None

    def create_if_needed(self, validity_start: str, release=1) -> None:
        """
        Initialize CKD file for binning tables if not exist

        Parameters
        ----------
        validity_start: str
           Validity start of the CKD data, as yyyymmddTHHMMSS
        release :  int, default=1
           Release number, start at 1
        """
        self.ckd_file = f'SPX1_CKD_BIN_TBL_{validity_start}_{release:03d}.nc'

        if (self.ckd_dir / self.ckd_file).is_file():
            return

        # initialize netCDF file with binning tables
        with Dataset(self.ckd_dir / self.ckd_file, 'w') as fid:
            fid.title = 'SPEXone Level-1 binning-tables'
            fid.Convensions = 'CF-1.6'
            fid.project = 'PACE Project'
            fid.instrument = 'SPEXone'
            fid.institution = 'SRON Netherlands Institute for Space Research'
            fid.processing_version = version.get()
            fid.validity_start = validity_start + '+00:00'
            fid.release_number = np.uint16(release)
            fid.date_created = datetime.now(timezone.utc).isoformat(
                timespec='seconds')

            fid.createDimension('row', 1024)
            fid.createDimension('column', 1024)

    def search(self, coverage_start=None) -> None:
        """
        Search CKD file with binning tables

        Parameters
        ----------
        coverage_start : str, default=None
           time_coverage_start or start of the measurement (UTC)

        Raises
        ------
        FileNotFoundError
           No CKD with binning tables found
        """
        ckd_files = list(Path(self.ckd_dir).glob('SPX1_CKD_BIN_TBL_*.nc'))
        if not ckd_files:
            raise FileNotFoundError('No CKD with binning tables found')
        ckd_files = [x.name for x in ckd_files]

        # use the latest version of the binning-table CKD
        if coverage_start is None:
            self.ckd_file = sorted(ckd_files)[-1]
            return

        # use binning-table CKD based on coverage_start
        coverage_date = datetime.fromisoformat(coverage_start)
        for ckd_fl in sorted(ckd_files, reverse=True):
            validity_date = datetime.strptime(ckd_fl.split('_')[4] + '+00:00',
                                              '%Y%m%dT%H%M%S%z')
            if validity_date < coverage_date:
                self.ckd_file = ckd_fl
                break
        else:
            raise FileNotFoundError('No valid CKD with binning tables found')

    def add_table(self, table_id: int, lineskip_arr, binning_table) -> None:
        """
        Add a binning table definition to existing CKD file

        Parameters
        ----------
        table_id :  int
          Table identifier (integer between 1 and 255)
        lineskip_arr :  ndarray
          Lineskip array definition
        binning_table :  ndarray
          Binning table definition
        """
        index, count = np.unique(binning_table[lineskip_arr == 1, :],
                                 return_counts=True)

        with Dataset(self.ckd_dir / self.ckd_file, 'r+') as fid:
            gid = fid.createGroup(f'/Table_{table_id:03d}')
            gid.enabled_lines = np.uint16(lineskip_arr.sum())
            gid.flex_binned_pixels = np.uint32(index.max()+1)
            gid.date_created = datetime.now(timezone.utc).isoformat(
                timespec='seconds')

            dset = gid.createVariable('binning_table', 'u4', ('row', 'column'),
                                      fill_value=FILL_VALUE,
                                      chunksizes=(128, 128),
                                      zlib=True, complevel=1, shuffle=True)
            dset.long_name = 'binning table'
            dset.valid_min = np.uint32(0)
            dset.valid_max = np.uint32(index.max())
            dset[:] = binning_table

            dset = gid.createVariable('lineskip_arr', 'u1', ('row',),
                                      zlib=True, complevel=1, shuffle=True)
            dset.long_name = 'lineskip array'
            dset.valid_min = np.uint8(0)
            dset.valid_max = np.uint8(1)
            dset[:] = lineskip_arr

            gid.createDimension('bins', count.size)
            dset = gid.createVariable('count_table', 'u2', ('bins',),
                                      zlib=True, complevel=1, shuffle=True)
            dset.long_name = 'number of aggregated pixel readings'
            dset.valid_min = np.uint16(0)
            dset.valid_max = np.uint16(count.max())
            dset[:] = count.astype('u2')

    def unbin(self, table_id: int, img_binned):
        """
        Return unbinned detector data

        Parameters
        ----------
        table_id :  int
           Table identifier (integer between 1 and 255)
        img_binned : ndarray
           Binned image data

        Returns
        -------
           Unbinned image data (no interpolation)
        """
        with Dataset(self.ckd_dir / self.ckd_file, 'r') as fid:
            gid = fid[f'/Table_{table_id:03d}']
            binning_table = gid.variables['binning_table'][:]
            lineskip_arr = gid.variables['lineskip_arr'][:]
            count_table = gid.variables['count_table'][:]

        revert = np.full(binning_table.shape, np.nan)
        table = binning_table[lineskip_arr == 1, :].reshape(-1)
        revert[lineskip_arr == 1, :] = \
            (img_binned / count_table)[table].reshape(-1, 1024)

        return revert
